from apisrun.utils.commons import (
    validate_path_file,
    validate_errors_apisrun_yml,
    load_yaml,
    has_open_port,
)
from apisrun.services.base import client
from apisrun.services.docker import service_docker
from apisrun.constants import (
    DockerConfig,
    DockerStatus,
    ApisrunAdmin,
    ApisrunErrorMsg,
    ApisrunMsg,
)
from rich.console import Console
from rich.table import Table


console = Console()


class ApisrunService:
    """Apisrun Service"""

    def _set_data_in_apisrun(self, data, port):
        url = ApisrunAdmin.NODE.format(port=port)
        errors = False
        for body in data:
            r = client.post(url, body, {})
            if r.status_code != 201:
                errors = True

        return errors

    def _wait_health_api(self, container, port):
        self._wait_container_status_expected(container)
        url = ApisrunAdmin.NODE.format(port=port)
        while True:
            resp = client.get(url, {})
            if resp:
                break

    def _wait_container_status_expected(
        self, container_name, status_expected=DockerStatus.RUNNING
    ):
        while True:
            container = service_docker.get_container(container_name)
            if container and container.status == status_expected:
                break

    def _get_available_port(self, initial_port: int):
        port = initial_port
        while True:
            if has_open_port(ApisrunAdmin.DEFAULT_HOST, port):
                break
            port = port + 1
        return port

    def _build_container_port(self, internal, external):
        return {f"{internal}/tcp": external}

    def _build_db_uri(self, api_name):
        container_name = DockerConfig.SERVICE_MONGO_NAME.format(api_name)
        return DockerConfig.APISRUN_MONGO_URI.format(
            name=container_name, port=DockerConfig.MONGO_PORT_DEFAULT
        )

    def _create_container_api(self, api_name, data):
        envs = {"ENGINE_URI": self._build_db_uri(api_name)}
        port = self._get_available_port(DockerConfig.APISRUN_PORT_DEFAULT)
        container_name = DockerConfig.SERVICE_NAME.format(api_name)

        console.print(ApisrunMsg.CREATING_API_CONTAINER.format(api_name))
        container = service_docker.run(
            DockerConfig.IMAGE_DEFAULT,
            container_name,
            DockerConfig.NETWORK_DEFAULT,
            envs,
            self._build_container_port(DockerConfig.APISRUN_PORT_DEFAULT, port),
            DockerConfig.LABELS_API,
        )
        if container:
            self._wait_container_status_expected(container_name)
            self._wait_health_api(container_name, port)
            console.print(ApisrunMsg.API_CONTAINER_DONE.format(api_name, port))
            console.print(ApisrunMsg.PUSH_DATA.format(api_name))
            errors = self._set_data_in_apisrun(data, port)
            if errors:
                console.print(ApisrunErrorMsg.PUSH_DATA_API_ERROR.format(api_name))
                return False
            return True

        console.print(ApisrunErrorMsg.CONTAINER_ERROR.format(api_name))
        return False

    def _create_container_db(self, api_name):
        port = self._get_available_port(DockerConfig.MONGO_PORT_DEFAULT)
        container_name = DockerConfig.SERVICE_MONGO_NAME.format(api_name)

        console.print(ApisrunMsg.CREATING_DB_CONTAINER.format(api_name))
        container = service_docker.run(
            DockerConfig.IMAGE_MONGO,
            container_name,
            DockerConfig.NETWORK_DEFAULT,
            DockerConfig.MONGO_ENVS,
            self._build_container_port(DockerConfig.MONGO_PORT_DEFAULT, port),
            DockerConfig.LABELS_DB,
        )
        if container:
            self._wait_container_status_expected(container_name)
            console.print(ApisrunMsg.DB_CONTAINER_DONE.format(api_name, port))
            return True

        console.print(ApisrunErrorMsg.CONTAINER_DB_ERROR.format(api_name))
        return False

    def _is_data_valid(self, path):
        path = ApisrunAdmin.FILE_DEFAULT if not path else path
        if not validate_path_file(path):
            console.print(ApisrunErrorMsg.FILE_NOT_FOUND.format(path))
            return None

        data = load_yaml(path)
        if validate_errors_apisrun_yml(data):
            console.print(ApisrunErrorMsg.VALIDATING_FILE.format(path))
            return None

        return data

    def create_node(self, path):
        """Create an node in apisrun."""
        data = self._is_data_valid(path)

        service_docker.get_or_create_network(DockerConfig.NETWORK_DEFAULT)

        for key, nodes in data.items():
            cdb = self._create_container_db(key)
            if not cdb:
                break
            capi = self._create_container_api(key, nodes)
            if not capi:
                break
            console.print(ApisrunMsg.SERVICE_DONE.format(key))

    def list_nodes(self):
        """List nodes in apisrun."""
        containers = service_docker.list(filters={"label": "apisrun"})
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Container")
        table.add_column("Status", justify="right")
        for c in containers:
            table.add_row(c.name, c.status)
        console.print(table)

    def kill_containers(self):
        """Stop containers."""
        containers = service_docker.list(filters={"label": "apisrun"})
        for c in containers:
            c.remove(force=True)
        console.print(ApisrunMsg.SERVICE_KILL)
